# pylint: disable=too-many-lines

"""
Various helpers.
"""

import collections
import contextlib
import errno
import functools
import io
import json
import os
import re
import shlex
import subprocess
import sys
import threading
import time
import warnings

import bs4
import urlnormalizer
import jinja2
import requests as original_requests

# Python 2/3 compatibility
import six
from six import PY2, ensure_str, iteritems, iterkeys
from six.moves import http_client, urllib

# Don't know why pylint reports "Relative import 'ruamel.yaml', should be 'gluetool.ruamel.yaml'" :(
# pylint: disable=relative-import
import ruamel.yaml

from .glue import GlueError, SoftGlueError, GlueCommandError
from .result import Result
from .log import Logging, ContextAdapter, PackageAdapter, LoggerMixin, BlobLogger, \
    log_blob, log_dict, print_wrapper

# Type annotations
# pylint: disable=unused-import, wrong-import-order
from typing import TYPE_CHECKING, cast  # noqa
from typing import Any, Callable, Deque, Dict, List, Optional, Pattern, Tuple, TypeVar, Union  # noqa
from .log import LoggingFunctionType  # noqa

if TYPE_CHECKING:
    import logging  # noqa


# Type variable used in generic types
T = TypeVar('T')


if PY2:
    DEVNULL = io.open(os.devnull, 'wb')

else:
    # In runtime, it is guarder by `PY2`, but Python 2 Pylint still checks it, and there's no `DEVNULL`
    # in Python 2... Hence disabling `no-name-in-module`.
    from subprocess import DEVNULL  # pylint: disable=ungrouped-imports,no-name-in-module


# Patch urlnormalizer to support file:// scheme.
if 'file' not in urlnormalizer.normalizer.SCHEMES:
    urlnormalizer.normalizer.SCHEMES = urlnormalizer.normalizer.SCHEMES + ('file',)


def deprecated(func):
    # type: (Callable[..., Any]) -> Callable[..., Any]

    """
    This is a decorator which can be used to mark functions as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        # type: (*Any, **Any) -> Any

        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn('Function {} is deprecated.'.format(func.__name__), category=DeprecationWarning,
                      stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter

        return func(*args, **kwargs)

    return new_func


def dict_update(dst, *args):
    # type: (Dict[Any, Any], *Dict[Any, Any]) -> Dict[Any, Any]

    """
    Python's ``dict.update`` does not return the dictionary just updated but a ``None``. This function
    is a helper that does updates the dictionary *and* returns it. So, instead of:

    .. code-block:: python

       d.update(other)
       return d

    you can use:

    .. code-block:: python

       return dict_update(d, other)

    :param dict dst: dictionary to be updated.
    :param args: dictionaries to update ``dst`` with.
    """

    for other in args:
        assert isinstance(other, dict)

        dst.update(other)

    return dst


def normalize_bool_option(option_value):
    # type: (Union[str, bool]) -> bool

    """
    Convert option value to Python's boolean.

    ``option_value`` is what all those internal option processing return,
    which may be a default value set for an option, or what user passed in.

    As switches, options with values can be used:

    .. code-block:: bash

       --foo=yes|no
       --foo=true|false
       --foo=1|0
       --foo=Y|N
       --foo=on|off

    With combination of ``store_true``/``store_false`` and a default value module developer
    sets for the option, simple form without value is evaluated as easily. With ``store_true``
    and ``False`` default, following option turn the feature `foo` on:

    .. code-block:: bash

       --enable-foo

    With ``store_false`` and ``True`` default, following simple option turn the feature `foo` off:

    .. code-block:: bash

       --disable-foo
    """

    if str(option_value).strip().lower() in ('yes', 'true', '1', 'y', 'on'):
        return True

    return False


def normalize_multistring_option(option_value, separator=','):
    # type: (Union[str, List[str]], Optional[str]) -> List[str]

    """
    Reduce string, representing comma-separated list of items, or possibly a list of such strings,
    to a simple list of items. Strips away the whitespace wrapping such items.

    .. code-block:: bash

        foo --option value1 --option value2, value3
        foo --option value1,value2,value3

    Or, when option is set by a config file:

    .. code-block:: bash

        option = value1
        option = value1, value2, \
                 value3

    After processing, different variants can be found when ``option('option')`` is called,
    ``['value1', 'value2,value3']``, ``['value1,value2,value3']``, ``'value1'`` and ``value1, value2, value3``.

    To reduce the necessary work, use this helper function to treat such option's value,
    and get simple ``['value1', 'value2', 'value3']`` structure.
    """

    if option_value is None:
        return []

    # If the value is string, convert it to list - it comes from a config file,
    # command-line parsing always produces a list. This reduces config file values
    # to the same structure command-line produces.
    values = [option_value] if isinstance(option_value, six.string_types) else option_value

    # Now deal with possibly multiple paths, separated by comma and some white space, inside
    # every item of the list. Split the paths in the item by the given separator, strip the
    # white space, and concatenate these lists (one for each item in the main `values` list)
    # using sum - with an empty list as a start, it works for lists just as nicely.
    return sum([
        [value.strip() for value in item.split(separator)] for item in values
    ], [])


def normalize_shell_option(option_value):
    # type: (Union[str, List[str]]) -> List[str]

    """
    Reduce string, using a shell-like syntax, or possibly a list of such strings,
    to a simple list of items. Strips away the whitespace wrapping such items.

    .. code-block:: bash

        foo --option value1 --option value2\\ value3 --option "value4 value5"

    Or, when option is set by a config file:

    .. code-block:: bash

        option = value1 value2\\ value3 "value4 value5"

    After processing, different variants can be found when ``option('option')`` is called,
    ``['value1', 'value2,value3']``, ``['value1,value2,value3']``, ``'value1'`` and ``value1, value2, value3``.

    To reduce the necessary work, use this helper function to treat such option's value,
    and get simple ``['value1', 'value2 value3', 'value4 value5']`` structure.
    """

    if not option_value:
        return []

    # If the value is string, convert it to list - it comes from a config file,
    # command-line parsing always produces a list. This reduces config file values
    # to the same structure command-line produces.
    values = [option_value] if isinstance(option_value, six.string_types) else option_value

    # Now split each item using shlex, and merge these lists into a single one.
    return sum([
        [ensure_str(s) for s in shlex.split(value)]
        for value in values
    ], [])


def normalize_path(path):
    # type: (str) -> str

    """
    Apply common treatments on a given path:

        * replace home directory reference (``~`` and similar), and
        * convert ``path`` to a normalized absolutized version of the pathname.
    """

    return os.path.abspath(os.path.expanduser(path))


def normalize_path_option(option_value, separator=','):
    # type: (Union[str, List[str]], Optional[str]) -> List[str]

    """
    Reduce many ways how list of paths is specified by user, to a simple list of paths. See
    :py:func:`normalize_multistring_option` for more details.
    """

    return [normalize_path(path) for path in normalize_multistring_option(option_value, separator=separator)]


class IncompatibleOptionsError(SoftGlueError):
    pass


class Bunch(object):
    # pylint: disable=too-few-public-methods

    @deprecated
    def __init__(self, **kwargs):
        # type: (**Any) -> None

        self.__dict__.update(kwargs)


class ThreadAdapter(ContextAdapter):
    """
    Custom logger adapter, adding thread name as a context.

    :param gluetool.log.ContextAdapter logger: parent logger whose methods will be used for logging.
    :param threading.Thread thread: thread whose name will be added.
    """

    def __init__(self, logger, thread):
        # type: (Union[logging.Logger, ContextAdapter], threading.Thread) -> None

        super(ThreadAdapter, self).__init__(logger, {'ctx_thread_name': (5, thread.name)})


class WorkerThread(LoggerMixin, threading.Thread):
    """
    Worker threads gets a job to do, and returns a result. It gets a callable, ``fn``,
    which will be called in thread's ``run()`` method, and thread's ``result`` property
    will be the result - value returned by ``fn``, or exception raised during the
    runtime of ``fn``.

    :param gluetool.log.ContextAdapter logger: logger to use for logging.
    :param fn: thread will start `fn` to do the job.
    :param fn_args: arguments for `fn`
    :param fn_kwargs: keyword arguments for `fn`
    """

    def __init__(self, logger, fn, fn_args=None, fn_kwargs=None, **kwargs):
        # type: (ContextAdapter, Callable[..., Any], Optional[Tuple[Any, ...]], Optional[Dict[str, Any]], **Any) -> None

        threading.Thread.__init__(self, **kwargs)
        LoggerMixin.__init__(self, ThreadAdapter(logger, self))

        self._fn = fn
        self._args = fn_args or ()
        self._kwargs = fn_kwargs or {}

        self.result = None  # type: Union[Exception, Any]

    def run(self):
        # type: () -> None

        self.debug('worker thread started')

        try:
            self.result = self._fn(*self._args, **self._kwargs)

        # pylint: disable=broad-except
        except Exception as e:
            self.error('exception raised in worker thread: {}'.format(e))
            self.result = e

        finally:
            self.debug('worker thread finished')


class StreamReader(object):
    def __init__(self, stream, name=None, block=16):
        # type: (Any, Optional[str], Optional[int]) -> None

        """
        Wrap blocking ``stream`` with a reading thread. The threads read from
        the (normal, blocking) `stream` and adds bits and pieces into the `queue`.
        ``StreamReader`` user then can check the `queue` for new data.
        """

        self._stream = stream
        self._name = name or stream.name

        # List would fine as well, however deque is better optimized for
        # FIFO operations, and it provides the same thread safety.
        self._queue = collections.deque()  # type: Deque[Union[None, str]]
        self._content = []  # type: List[str]

        def _enqueue():
            # type: () -> None

            """
            Read what's available in stream and add it into the queue
            """

            while True:
                data = self._stream.read(block)

                if not data:
                    # signal EOF
                    self._queue.append('')
                    return

                self._queue.append(data)
                self._content.append(data)

        self._thread = threading.Thread(target=_enqueue)
        self._thread.daemon = True
        self._thread.start()

    @property
    def name(self):
        # type: () -> str

        return self._name

    @property
    def content(self):
        # type: () -> str

        return ''.join(self._content)

    def wait(self):
        # type: () -> None

        self._thread.join()

    def read(self):
        # type: () -> Optional[str]

        try:
            return cast(str, self._queue.popleft())

        except IndexError:
            return None


class ProcessOutput(object):
    """
    Result of external process.
    """

    # pylint: disable=too-many-arguments,too-few-public-methods
    def __init__(self, cmd, exit_code, stdout, stderr, kwargs):
        # type: (List[str], int, Optional[str], Optional[str], Dict[str, Any]) -> None

        self.cmd = cmd
        self.kwargs = kwargs

        self.exit_code = exit_code
        self.stdout = stdout
        self.stderr = stderr

    def log_stream(self, stream, logger):
        # type: (str, ContextAdapter) -> None

        content = getattr(self, stream)

        if content is None:
            if stream in self.kwargs:
                logger.debug('{}:\n  command produced no output'.format(stream))
            else:
                logger.debug('{}:\n  command forwarded the output to its parent'.format(stream))

        else:
            log_blob(logger.verbose, stream, content)

    def log(self, logger):
        # type: (ContextAdapter) -> None

        logger.debug('command exited with code {}'.format(self.exit_code))

        self.log_stream('stdout', logger)
        self.log_stream('stderr', logger)


class Command(LoggerMixin, object):
    """
    Wrap an external command, its options and other information, necessary for running the command.

    The main purpose is to gather all relevant pieces into a single space, call :py:class:`subprocess.Popen`,
    and log everything.

    By default, both standard output and error output are of the process are captured and returned back to
    caller. Under some conditions, caller might want to see the output in "real-time". For that purpose,
    they can pass callable via ``inspect_callback`` parameter - such callable will be called for every received
    bit of input on both standard and error outputs. E.g.

    .. code-block:: python

       def foo(stream, s, flush=False):
           if s is not None and 'a' in s:
               print s

       Command(['/bin/foo']).run(inspect=foo)

    This example will print all substrings containing letter `a`. Strings passed to ``foo`` may be of arbitrary
    lengths, and may change between subsequent use of ``Command`` class.

    :param list executable: Executable to run. Feel free to use the whole command, including its options,
        if you have no intention to modify them before running the command.
    :param list options: If set, it's a list of options to pass to the ``executable``. Options are
        specified in a separate list to allow modifications of ``executable`` and ``options``
        before actually running the command.
    :param gluetool.log.ContextAdapter logger: Parent logger whose methods will be used for logging.


    .. versionadded:: 1.1
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, executable, options=None, logger=None):
        # type: (List[str], Optional[List[str]], Optional[ContextAdapter]) -> None

        super(Command, self).__init__(logger or Logging.get_logger())

        self.executable = executable
        self.options = options or []  # type: List[str]

        self.use_shell = False
        self.quote_args = False

        self._command = None  # type: Optional[List[str]]
        self._popen_kwargs = None  # type: Optional[Dict[str, Any]]
        self._process = None  # type: Optional[subprocess.Popen]
        self._exit_code = None  # type: Optional[int]

        self._stdout = None  # type: Optional[str]
        self._stderr = None  # type: Optional[str]

    def _apply_quotes(self):
        # type: () -> List[str]

        """
        Return options to pass to ``Popen``. Applies quotes as necessary.
        """

        if not self.quote_args:
            return self.executable + self.options

        # escape apostrophes in strings and adds them around strings with space
        return [('"{}"'.format(option.replace('"', r'\"')) if ' ' in option and not
                 (
                     (option.startswith('"') and option.endswith('"')) or
                     (option.startswith("'") and option.endswith("'")))
                 else option) for option in self.executable + self.options]

    def _communicate_batch(self):
        # type: () -> None

        # Collapse optionals to specific types
        assert self._process is not None

        self._stdout, self._stderr = self._process.communicate()

    def _communicate_inspect(self, inspect_callback):
        # type: (Optional[Callable[[Any, Optional[str], bool], None]]) -> None

        # Collapse optionals to specific types
        assert self._command is not None
        assert self._process is not None

        # let's capture *both* streams - capturing just a single one leads to so many ifs
        # and elses and messy code
        p_stdout = StreamReader(self._process.stdout, name='<stdout>')
        p_stderr = StreamReader(self._process.stderr, name='<stderr>')

        if inspect_callback is None:
            def stdout_write(stream, data, flush):
                # type: (Any, Optional[str], bool) -> None

                # pylint: disable=unused-argument

                if data is None:
                    return

                # Not suitable for multiple simultaneous commands. Shuffled output will
                # ruin your day. And night. And few following weeks, full of debugging, as well.
                sys.stdout.write(ensure_str(data))
                sys.stdout.flush()

            inspect_callback = stdout_write

        inputs = (p_stdout, p_stderr)

        with BlobLogger('Output of command: {}'.format(format_command_line([self._command])),
                        outro='End of command output',
                        writer=self.info):
            self.debug("output of command is inspected by the caller")
            self.debug('following blob-like header and footer are expected to be empty')
            self.debug('the captured output will follow them')

            # As long as process runs, keep calling callbacks with incoming data
            while True:
                for stream in inputs:
                    inspect_callback(stream, stream.read(), False)

                if self._process.poll() is not None:
                    break

                # give up OS' attention and let others run
                # time.sleep(0) is a Python synonym for "thread yields the rest of its quantum"
                time.sleep(0.1)

            # OK, process finished but we have to wait for our readers to finish as well
            p_stdout.wait()
            p_stderr.wait()

            for stream in inputs:
                while True:
                    data = stream.read()

                    if data in ('', None):
                        break

                    inspect_callback(stream, data, False)

                inspect_callback(stream, None, True)

        self._stdout, self._stderr = p_stdout.content, p_stderr.content

    def _construct_output(self):
        # type: () -> ProcessOutput

        # Collapse optionals to specific types
        assert self.logger is not None
        assert self._command is not None
        assert self._exit_code is not None
        assert self._popen_kwargs is not None

        output = ProcessOutput(self._command, self._exit_code, self._stdout, self._stderr, self._popen_kwargs)

        output.log(self.logger)

        return output

    def run(self, inspect=False, inspect_callback=None, **kwargs):
        # type: (Optional[bool], Optional[Callable[..., None]], **Any) -> ProcessOutput

        """
        Run the command, wait for it to finish and return the output.

        :param bool inspect: If set, ``inspect_callback`` will receive the output of command in "real-time".
        :param callable inspect_callback: callable that will receive command output. If not set, default
            "write to ``sys.stdout``" is used.
        :rtype: gluetool.utils.ProcessOutput instance
        :returns: :py:class:`gluetool.utils.ProcessOutput` instance whose attributes contain data returned
            by the child process.
        :raises gluetool.glue.GlueError: When somethign went wrong.
        :raises gluetool.glue.GlueCommandError: When command exited with non-zero exit code.
        """

        # pylint: disable=too-many-branches

        def _check_types(items):
            # type: (List[str]) -> None

            if not isinstance(items, list):
                raise GlueError('Only list of strings is accepted')

            if not all((isinstance(s, six.string_types) for s in items)):
                raise GlueError('Only list of strings is accepted, {} found'.format([(s, type(s)) for s in items]))

        _check_types(self.executable)
        _check_types(self.options)

        self._command = self._apply_quotes()

        if self.use_shell is True:
            self._command = [' '.join(self._command)]

        # Set default stdout/stderr, unless told otherwise
        if 'stdout' not in kwargs:
            kwargs['stdout'] = subprocess.PIPE

        if 'stderr' not in kwargs:
            kwargs['stderr'] = subprocess.PIPE

        if self.use_shell:
            kwargs['shell'] = True

        self._popen_kwargs = kwargs

        def _format_stream(stream):
            # type: (Any) -> str

            if stream == subprocess.PIPE:
                return 'PIPE'
            if stream == DEVNULL:
                return 'DEVNULL'
            if stream == subprocess.STDOUT:
                return 'STDOUT'
            return str(stream)

        printable_kwargs = kwargs.copy()  # type: Dict[str, Any]
        for stream in ('stdout', 'stderr'):
            if stream in printable_kwargs:
                printable_kwargs[stream] = _format_stream(printable_kwargs[stream])

        log_dict(self.debug, 'command', self._command)
        log_dict(self.debug, 'kwargs', printable_kwargs)
        log_blob(self.debug, 'runnable (copy & paste)', format_command_line([self._command]))

        try:
            self._process = subprocess.Popen(self._command, **self._popen_kwargs)

            if inspect is True:
                self._communicate_inspect(inspect_callback)

            else:
                self._communicate_batch()

        except OSError as e:
            if e.errno == errno.ENOENT:
                raise GlueError("Command '{}' not found".format(ensure_str(self._command[0])))

            raise e

        self._exit_code = self._process.poll()

        output = self._construct_output()

        if self._exit_code != 0:
            raise GlueCommandError(self._command, output)

        return output


@deprecated
def run_command(cmd, logger=None, inspect=False, inspect_callback=None, **kwargs):
    # type: (List[str], Optional[ContextAdapter], bool, Optional[Callable[..., None]], **Any) -> ProcessOutput

    # pylint: disable=unused-argument

    """
    Wrapper for ``Command(...).run().

    Provided for backward compatibility.

    .. deprecated:: 1.1

        Use :py:class:`gluetool.utils.Command` instead.
    """

    return Command(cmd, logger=logger).run(**kwargs)


def check_for_commands(cmds):
    # type: (List[str]) -> None

    """ Checks if all commands in list cmds are valid """

    for cmd in cmds:
        try:
            Command(['/bin/bash', '-c', 'command -v {}'.format(cmd)]).run(stdout=DEVNULL)

        except GlueError:
            raise GlueError("Command '{}' not found on the system".format(ensure_str(cmd)))


class cached_property(object):
    # pylint: disable=invalid-name,too-few-public-methods
    """
    ``property``-like decorator - at first access, it calls decorated
    method to acquire the real value, and then replaces itself with
    this value, making it effectively "cached". Useful for properties
    whose value does not change over time, and where getting the real
    value could penalize execution with unnecessary (network, memory)
    overhead.

    Delete attribute to clear the cached value - on next access, decorated
    method will be called again, to acquire the real value.

    Of possible options, only read-only instance attribute access is
    supported so far.
    """

    def __init__(self, method):
        # type: (Callable[..., Any]) -> None

        self._method = method
        self.__doc__ = getattr(method, '__doc__')

    def __get__(self, obj, cls):
        # type: (Any, Any) -> Any

        # does not support class attribute access, only instance
        assert obj is not None

        # get the real value of this property
        value = self._method(obj)

        # replace cached_property instance with the value
        obj.__dict__[self._method.__name__] = value

        return value


def format_command_line(cmdline):
    # type: (List[List[str]]) -> str

    """
    Return formatted command-line.

    All but the first line are indented by 4 spaces.

    :param list cmdline: list of iterables, representing command-line split to multiple lines.
    """

    def _format_options(options):
        # type: (List[str]) -> str

        # To make code more readable, it's split to multiple lines. First, make sure each option
        # is "str", accepted by `shlex_quote` function.
        encoded_options = [ensure_str(opt) for opt in options]

        # Next, quote each option.
        # shlex_quote takes one argument, pylint thinks otherwise :/
        # pylint: disable=too-many-function-args
        quoted_options = [six.moves.shlex_quote(opt) for opt in encoded_options]

        # Finally, convert quoted options back to "text".
        decoded_options = [ensure_str(opt) for opt in quoted_options]

        return ' '.join(decoded_options)

    cmd = [_format_options(cmdline[0])]

    for row in cmdline[1:]:
        cmd.append('    ' + _format_options(row))

    return '\n'.join(cmd)


@deprecated
def fetch_url(url, logger=None, success_codes=(200,)):
    # type: (str, Optional[ContextAdapter], Tuple[int, ...]) -> Tuple[Any, str]

    """
    "Get me content of this URL" helper.

    Very thin wrapper around urllib. Added value is logging, and converting
    possible errors to :py:class:`gluetool.glue.GlueError` exception.

    :param url: URL to get.
    :param gluetool.log.ContextLogger logger: Logger used for logging.
    :param tuple success_codes: tuple of HTTP response codes representing successfull request.
    :returns: tuple ``(response, content)`` where ``response`` is what
      :py:func:`requests.get` returns, and ``content`` is the payload
      of the response.
    """

    logger = logger or Logging.get_logger()

    logger.debug("opening URL '{}'".format(url))

    try:
        with requests(logger=logger) as req:
            response = req.get(url)

    except urllib.error.HTTPError as exc:
        raise GlueError("Failed to fetch URL '{}': {}".format(url, exc))

    if response.status_code not in success_codes:
        raise GlueError("Unsuccessfull response from '{}'".format(url))

    return response, response.content


@contextlib.contextmanager
def requests(logger=None):
    # type: (Optional[ContextAdapter]) -> Any

    """
    Wrap :py:mod:`requests` with few layers providing us with the logging and better insight into
    what has been happening when ``requests`` did their job.

    Used as a context manager, yields a patched ``requests`` module. As long as inside the context,
    detailed information about HTTP traffic are logged via given logger.

    .. note::

       The original ``requests`` library is returned, with slight modifications for better integration
       with ``gluetool`` logging facilities. Each and every ``requests`` API feature is available
       and , hopefully, enhancements applied by this wrapper wouldn't interact with ``requests``
       functionality.

    .. code-block:: python

       with gluetool.utils.requests() as R:
           R.get(...).json()
           ...

           r = R.post(...)
           assert r.code == 404
           ...

    :param logger: used for logging.
    :returns: :py:mod:`requests` module.
    """

    # Enable http_client debugging. It's being used underneath ``requests`` and ``urllib3``,
    # but it's stupid - uses "print" instead of a logger, therefore we have to capture it
    # and disable debug logging when leaving the context.
    logger = logger or Logging.get_logger()

    http_client_logger = PackageAdapter(logger, 'httplib') if PY2 else PackageAdapter(logger, 'http_client')
    http_client.HTTPConnection.debuglevel = 1  # type: ignore

    # Start capturing ``print`` statements - they are used to provide debug messages, therefore
    # using ``debug`` level.
    with print_wrapper(log_fn=http_client_logger.debug):  # type: ignore  # logger.debug signature is compatible
        # To log responses and their content, we must take a look at ``Response`` instance
        # returned by several entry methods (``get``, ``post``, ...). To do that, we have
        # a simple wrapper function.

        # ``original_method`` is the actual ``requests.foo`` (``get``, ``post``, ...), wrapper
        # calls it to do the job, and logs response when it's done.
        def _verbose_request(original_method, *args, **kwargs):
            # type: (Callable[..., Any], *Any, **Any) -> Any

            ret = original_method(*args, **kwargs)

            assert logger is not None
            log_blob(logger.debug,
                     'response content',
                     ret.content)

            return ret

        # gather the original methods...
        methods = {
            method_name: getattr(original_requests, method_name)
            for method_name in ('head', 'get', 'post', 'put', 'patch', 'delete')
        }

        # ... and replace them with our wrapper, giving it the original method as the first argument
        for method_name, original_method in iteritems(methods):
            setattr(original_requests, method_name, functools.partial(_verbose_request, original_method))

        try:
            yield original_requests

        finally:
            # put original methods back...
            for method_name, original_method in iteritems(methods):
                setattr(original_requests, method_name, original_method)

            # ... and disable http_client debugging
            http_client.HTTPConnection.debuglevel = 0  # type: ignore


def treat_url(url, logger=None):
    # type: (str, Optional[ContextAdapter]) -> str

    """
    Remove "weird" artifacts from the given URL. Collapse adjacent '.'s, apply '..', etc.

    :param text url: URL to clear.
    :param gluetool.log.ContextAdapter logger: logger to use for logging.
    :rtype: text
    :raises: gluetool.glue.GlueError: if URL is invalid
    :returns: Treated URL.
    """

    logger = logger or Logging.get_logger()

    logger.debug("treating a URL '{}'".format(url))

    norm_url = urlnormalizer.normalize_url(url)

    if norm_url is None:
        raise GlueError("'{}' does not look like an URL".format(url))

    return ensure_str(norm_url.strip())


def render_template(template, logger=None, **kwargs):
    # type: (Union[str, jinja2.environment.Template], Optional[ContextAdapter], **Any) -> str

    """
    Render Jinja2 template. Logs errors, and raises an exception when it's not possible
    to correctly render the template.

    :param template: Template to render. It can be either :py:class:`jinja2.environment.Template` instance,
        or a string.
    :param dict kwargs: Keyword arguments passed to render process.
    :returns: Rendered template.
    :raises gluetool.glue.GlueError: when the rednering failed.
    """

    logger = logger or Logging.get_logger()

    assert logger is not None

    try:
        def _render(template, source):
            # type: (jinja2.Template, str) -> str

            assert logger is not None

            log_blob(logger.debug, 'rendering template', source)
            log_dict(logger.verbose, 'context', kwargs)

            return ensure_str(template.render(**kwargs).strip())

        if isinstance(template, six.string_types):
            return _render(jinja2.Template(template), template)

        if isinstance(template, jinja2.environment.Template):
            if template.filename != '<template>':  # type: ignore
                with io.open(template.filename, 'r', encoding='utf-8') as f:  # type: ignore  # .filename attr exists
                    return _render(template, f.read())

            return _render(template, '<unknown template source>')

        raise GlueError('Unhandled template type {}'.format(type(template)))

    except Exception as exc:
        raise GlueError('Cannot render template: {}'.format(exc))


# pylint: disable=invalid-name
def YAML():
    # type: () -> ruamel.yaml.YAML

    """
    Provides YAML read/write interface with common settings.

    :rtype: ruamel.yaml.YAML
    """

    yaml = ruamel.yaml.YAML()
    yaml.indent(sequence=4, mapping=4, offset=2)

    return yaml


def from_yaml(yaml_string):
    # type: (str) -> Any

    """
    Convert YAML in a string into Python data structures.

    Uses internal YAML parser to produce result. Paired with :py:func:`load_yaml` and their
    JSON siblings to provide unified access to JSON and YAML.
    """

    return YAML().load(yaml_string)


def load_yaml(filepath, logger=None):
    # type: (str, Optional[ContextAdapter]) -> Any

    """
    Load data stored in YAML file, and return their Python representation.

    :param text filepath: Path to a file. ``~`` or ``~<username>`` are expanded before using.
    :param gluetool.log.ContextLogger logger: Logger used for logging.
    :rtype: object
    :returns: structures representing data in the file.
    :raises gluetool.glue.GlueError: if it was not possible to successfully load content of the file.
    """

    if not filepath:
        raise GlueError('File path is not valid: {}'.format(filepath))

    logger = logger or Logging.get_logger()

    real_filepath = normalize_path(filepath)

    logger.debug("attempt to load YAML from '{}' (maps to '{}')".format(filepath, real_filepath))

    if not os.path.exists(real_filepath):
        raise GlueError("File '{}' does not exist".format(filepath))

    try:
        with open(real_filepath, 'r') as f:
            data = YAML().load(f)

        log_dict(logger.debug, "loaded YAML data from '{}'".format(filepath), data)

        return data

    except ruamel.yaml.YAMLError as e:
        raise GlueError("Unable to load YAML file '{}': {}".format(filepath, e))


def dump_yaml(data, filepath, logger=None):
    # type: (Any, str, Optional[ContextAdapter]) -> None

    """
    Save data stored in variable to YAML file.

    :param object data: Data to store in YAML file
    :param text filepath: Path to an output file.
    :raises gluetool.glue.GlueError: if it was not possible to successfully save data to file.
    """
    if not filepath:
        raise GlueError("File path is not valid: '{}'".format(filepath))

    logger = logger or Logging.get_logger()

    real_filepath = normalize_path(filepath)
    dirpath = os.path.dirname(real_filepath)

    if not os.path.exists(dirpath):
        raise GlueError("Cannot save file in nonexistent directory '{}'".format(dirpath))

    try:
        with open(real_filepath, 'w') as f:
            YAML().dump(data, f)
            f.flush()

    except ruamel.yaml.YAMLError as e:
        raise GlueError("Unable to save YAML file '{}': {}".format(filepath, e))


def _json_byteify(data, ignore_dicts=False):
    # type: (Any, Optional[bool]) -> Any

    # if this is a unicode string, return it
    if isinstance(data, six.string_types):
        return data

    # if this is a list of values, return list of byteified values
    if isinstance(data, list):
        return [_json_byteify(item, ignore_dicts=True) for item in data]

    # if this is a dictionary, return dictionary of byteified keys and values
    # but only if we haven't already byteified it
    if isinstance(data, dict) and not ignore_dicts:
        return {
            _json_byteify(key, ignore_dicts=True): _json_byteify(value, ignore_dicts=True)
            for key, value in iteritems(data)
        }

    # if it's anything else, return it in its original form
    return data


def from_json(json_string):
    # type: (str) -> Any

    """
    Convert JSON in a string into Python data structures.

    Similar to :py:func:`json.loads` but uses special object hook to avoid unicode strings
    in the output..
    """

    return _json_byteify(json.loads(json_string, object_hook=_json_byteify), ignore_dicts=True)


def load_json(filepath, logger=None):
    # type: (str, Optional[ContextAdapter]) -> Any

    """
    Load data stored in JSON file, and return their Python representation.

    :param text filepath: Path to a file. ``~`` or ``~<username>`` are expanded before using.
    :param gluetool.log.ContextLogger logger: Logger used for logging.
    :rtype: object
    :returns: structures representing data in the file.
    :raises gluetool.glue.GlueError: if it was not possible to successfully load content of the file.
    """

    if not filepath:
        raise GlueError('File path is not valid: {}'.format(filepath))

    logger = logger or Logging.get_logger()

    real_filepath = normalize_path(filepath)

    logger.debug("attempt to load JSON from '{}' (maps to '{}')".format(filepath, real_filepath))

    if not os.path.exists(real_filepath):
        raise GlueError("File '{}' does not exist".format(filepath))

    try:
        with open(real_filepath, 'r') as f:
            data = _json_byteify(json.load(f, object_hook=_json_byteify), ignore_dicts=True)
            log_dict(logger.debug, "loaded JSON data from '{}'".format(filepath), data)

            return data

    except Exception as exc:
        raise GlueError("Unable to load JSON file '{}': {}".format(filepath, exc))


def _load_yaml_variables(data, enabled=True, logger=None):
    # type: (Any, bool, Optional[ContextAdapter]) -> Callable[[str], Union[str, List[str]]]
    """
    Load all variables from files referenced by a YAML, and return function to render a string
    as a template using these variables. The files containing variables are mentioned in comments,
    in a form ``# !include <filepath>`` form.

    :param data: data loaded from a YAML file.
    :param bool enabled: when set to ``False``, variables are not loaded and a simple no-op
        function is returned.
    :param gluetool.log.ContextLogger logger: Logger used for logging.
    :returns: Function accepting a string and returning a rendered template.
    """

    logger = logger or Logging.get_logger()

    def _render_template_nop(s):
        # type: (str) -> str

        return s

    if not enabled:
        return _render_template_nop

    # Our YAML reader preserves comments, they are accessible via `ca` attribute of the
    # YAML data structure (which behaves like a dict or list, but it has additional
    # powers).
    if not hasattr(data, 'ca') or not hasattr(data.ca, 'comment') or len(data.ca.comment) <= 1:
        logger.debug('when looking for !import directives, no comments found in YAML data')

        return _render_template_nop

    # Ok, so this YAML data contains comments. Check their values to find `!include` directives.
    # Load referenced files and merged them into a single context.
    context = {}  # type: Dict[str, Any]

    for comment in data.ca.comment[1]:
        value = comment.value.strip()

        if not value.startswith('# !include'):
            continue

        try:
            variables_map_path = shlex.split(value[2:])[1]

        except Exception as exc:
            raise GlueError("Cannot extract filename to include from '{}': {}".format(value, exc))

        logger.debug("loading variables from '{}'".format(variables_map_path))

        context.update(load_yaml(variables_map_path, logger=logger))

    def _render_template(s):
        # type: (Union[str, List[str]]) -> Union[str, List[str]]

        if isinstance(s, six.string_types):
            return render_template(s, logger=logger, **context)

        if isinstance(s, list):
            return [
                render_template(t, logger=logger, **context)
                for t in s
            ]

        raise GlueError("Don't know how to render object of type {}".format(type(s)))

    return _render_template


class SimplePatternMap(LoggerMixin, object):
    # pylint: disable=too-few-public-methods

    """
    `Pattern map` is a list of ``<pattern>``: ``<result>`` pairs. ``Pattern`` is a
    regular expression used to match a string, ``result`` is what the matching
    string maps to.

    Basically an ordered dictionary with regexp matching of keys, backed by an YAML file.

    :param str filepath: Path to a YAML file with map definition.
    :param gluetool.log.ContextLogger logger: Logger used for logging.
    :param bool allow_variables: if set, both patterns and converters are first treated as templates,
        and as such are rendered before doing anything else. Map may contain special comments,
        ``# !include <path>``, where path refers to a YAML file providing the necessary variables.
    """

    def __init__(self, filepath, logger=None, allow_variables=False):
        # type: (str, Optional[ContextAdapter], bool) -> None

        super(SimplePatternMap, self).__init__(logger or Logging.get_logger())

        pattern_map = load_yaml(filepath, logger=self.logger)

        if pattern_map is None:
            raise GlueError("pattern map '{}' does not contain any patterns".format(filepath))

        _render_template = _load_yaml_variables(pattern_map, enabled=allow_variables, logger=self.logger)

        self._compiled_map = []  # type: List[Tuple[Pattern[str], str]]

        for pattern_dict in pattern_map:
            if not isinstance(pattern_dict, dict):
                raise GlueError("Invalid format: '- <pattern>: <result>' expected, '{}' found".format(pattern_dict))

            pattern = next(iterkeys(pattern_dict))
            result = pattern_dict[pattern].strip()

            # Apply variables if requested.
            pattern = _render_template(pattern)
            result = _render_template(result)

            log_dict(logger.debug,  # type: ignore  # logger.debug signature is compatible
                     "rendered mapping '{}'".format(pattern),
                     result)

            try:
                pattern = re.compile(pattern)

            except re.error as exc:
                raise GlueError("Pattern '{}' is not valid: {}".format(pattern, exc))

            self._compiled_map.append((pattern, result))

    def match(self, s):
        # type: (str) -> str

        """
        Try to match ``s`` by the map. If the match is found - the first one wins - then its
        transformation is applied to the ``s``.

        :rtype: text
        :returns: if matched, output of the corresponding transformation.
        """

        self.debug("trying to match string '{}' with patterns in the map".format(s))

        for pattern, result in self._compiled_map:
            self.debug("testing pattern '{}'".format(pattern.pattern))

            match = pattern.match(s)
            if match is None:
                continue

            self.debug('  matched!')
            return result

        raise GlueError("Could not match string '{}' with any pattern".format(s))


class PatternMap(LoggerMixin, object):
    # pylint: disable=too-few-public-methods

    """
    `Pattern map` is a list of ``<pattern>``: ``<converter>`` pairs. ``Pattern`` is a
    regular expression used to match a string, ``converter`` is a function that transforms
    a string into another one, accepting the pattern and the string as arguments.

    It is defined in a YAML file:

    .. code-block:: yaml

       ---
       - 'foo-(\\d+)': 'bar-\\1'
       - 'baz-(\\d+)': 'baz, find_the_most_recent, append_dot'
       - 'bar-(\\d+)':
         - 'bar, find_the_most_recent, append_dot'
         - 'bar, find_the_oldest, append_dot'

    Patterns are the keys in each pair, while ``converter`` is a string (or list of strings),
    consisting of multiple items, separated by comma. The first item is **always** a string,
    let's call it ``R``. ``R``, given input string ``S1`` and the pattern, is used to transform
    ``S1`` to a new string, ``S2``, by calling ``pattern.sub(R, S1)``. ``R`` can make use of anything
    :py:meth:`re.sub` supports, including capturing groups.

    If there are other items in the ``converter`` string, they are names of `spices`, additional
    functions that will be called with ``pattern`` and the output of the previous spicing function,
    starting with ``S2`` in the case of the first `spice`.

    To allow spicing, user of ``PatternMap`` class must provide `spice makers` - mapping between
    `spice` names and functions that generate spicing functions. E.g.:

    .. code-block:: python

       def create_spice_append_dot(previous_spice):
           def _spice(pattern, s):
               s = previous_spice(pattern, s)
               return s + '.'
           return _spice

    ``create_spice_append_dot`` is a `spice maker`, used during creation of a pattern map after
    its definition is read, ``_spice`` is the actual spicing function used during the transformation
    process.

    There can be multiple converters for a single pattern, resulting in multiple values returned
    when the input string matches the corresponding pattern.

    :param text filepath: Path to a YAML file with map definition.
    :param dict spices: apping between `spices` and their `makers`.
    :param gluetool.log.ContextLogger logger: Logger used for logging.
    :param bool allow_variables: if set, both patterns and converters are first treated as templates,
        and as such are rendered before doing anything else. Map may contain special comments,
        ``# !include <path>``, where path refers to a YAML file providing the necessary variables.
    """

    def __init__(self,
                 filepath,  # type: str
                 spices=None,  # type: Optional[Dict[str, Callable[..., Callable[[Any, str], str]]]]
                 logger=None,  # type: Optional[ContextAdapter]
                 allow_variables=False  # type: bool
                ):  # noqa
        # type: (...) -> None

        super(PatternMap, self).__init__(logger or Logging.get_logger())

        spices = spices or {}

        pattern_map = load_yaml(filepath, logger=self.logger)

        if pattern_map is None:
            raise GlueError("pattern map '{}' does not contain any patterns".format(filepath))

        _render_template = _load_yaml_variables(pattern_map, enabled=allow_variables, logger=self.logger)

        def _create_simple_repl(repl):
            # type: (str) -> Callable[[Pattern[str], str], str]

            def _replace(pattern, target):
                # type: (Pattern[str], str) -> Any

                """
                Use `repl` to construct image from `target`, honoring all backreferences made by `pattern`.
                """

                self.debug("pattern '{}', repl '{}', target '{}'".format(pattern.pattern, repl, target))

                try:
                    return pattern.sub(repl, target)

                except re.error as e:
                    raise GlueError("Cannot transform pattern '{}' with target '{}', repl '{}': {}".format(
                        pattern.pattern, target, repl, e))

            return _replace

        self._compiled_map = []  # type: List[Tuple[Pattern[str], List[Callable[[Pattern[str], str], str]]]]

        for pattern_dict in pattern_map:
            log_dict(self.debug, 'pattern dict', pattern_dict)

            if not isinstance(pattern_dict, dict):
                raise GlueError("Invalid format: '- <pattern>: <transform>' expected, '{}' found".format(pattern_dict))

            # There is always just a single key, the pattern.
            pattern_key = next(iterkeys(pattern_dict))

            # Apply variables if requested.
            pattern = _render_template(pattern_key)
            converter_chains = _render_template(pattern_dict[pattern_key])

            # Given how YAML works, `pattern` is a string, but the type of `_render_template` return value
            # is Union[str, List[str]] - this covers possible lists on the right side of the equation.
            # To make mypy happy, let's collapse type of `pattern`.
            assert isinstance(pattern, six.string_types)

            log_dict(logger.debug,  # type: ignore  # logger.debug signature is compatible
                     "rendered mapping '{}'".format(pattern),
                     converter_chains)

            if isinstance(converter_chains, six.string_types):
                converter_chains = [converter_chains]

            try:
                compiled_pattern = re.compile(pattern)

            except re.error as e:
                raise GlueError("Pattern '{}' is not valid: {}".format(pattern, e))

            compiled_chains = []

            for chain in converter_chains:
                converters = [s.strip() for s in chain.split(',')]

                # first item in `converters` is always a simple string used by `pattern.sub()` call
                converter = _create_simple_repl(converters.pop(0))

                # if there any any items left, they name "spices" to apply, one by one,
                # on the result of the first operation
                for spice in converters:
                    if spice not in spices:
                        raise GlueError("Unknown 'spice' function '{}'".format(spice))

                    converter = spices[spice](converter)

                compiled_chains.append(converter)

            self._compiled_map.append((compiled_pattern, compiled_chains))

    def match(self, s, multiple=False):
        # type: (str, bool) -> Union[str, List[str]]

        """
        Try to match ``s`` by the map. If the match is found - the first one wins - then its
        conversions are applied to the ``s``.

        There can be multiple conversions for a pattern, by default only the product of
        the first one is returned. If ``multiple`` is set to ``True``, list of all products
        is returned instead.

        :rtype: text
        :returns: if matched, output of the corresponding transformation.
        """

        self.debug("trying to match string '{}' with patterns in the map".format(s))

        for pattern, converters in self._compiled_map:
            self.debug("testing pattern '{}'".format(pattern.pattern))

            match = pattern.match(s)
            if match is None:
                continue

            self.debug('  matched!')

            if multiple is not True:
                return converters[0](pattern, s)

            return [
                converter(pattern, s) for converter in converters
            ]

        raise GlueError("Could not match string '{}' with any pattern".format(s), sentry_fingerprint=[s])


# The type of `wait` is a bit complicated: callback returns `Result` instance with either valid value A, or
# error B. `wait` us supposed to return `A` when finished successfully - it raises an timeout exception otherwise
# anyway, so there's no returning of B. So, given that `check` returns value of type `Result[T, E]`, we can
# deduce that `wait` must return value of type T. The trick is to use `T` in both `WaitCheckType` and
# `wait` signature.

#: Generic type for callbacks used by :py:func:`wait` function. Accepts no arguments, returns an instance
#: of :py:class:`gluetool.result.Result`.
WaitCheckType = Callable[[], Result[T, Any]]


def wait(label, check, timeout=None, tick=30, logger=None):
    # type: (str, WaitCheckType[T], Optional[int], int, Optional[ContextAdapter]) -> T
    """
    Wait for a condition to be true.

    :param text label: printable label used for logging.
    :param callable check: called to test the condition. It must be of type :py:data:`WaitCheckType`: takes
        no arguments, must return instance of :py:class:`gluetool.Result`. If the result is valid, the condition
        is assumed to pass the check and waiting ends.
    :param int timeout: fail after this many seconds. ``None`` means test forever.
    :param int tick: test condition every ``tick`` seconds.
    :param gluetool.log.ContextAdapter logger: parent logger whose methods will be used for logging.
    :raises gluetool.glue.GlueError: when ``timeout`` elapses while condition did not pass the check.
    :returns: if the condition became true, the value returned by the ``check`` function
        is returned. It is unpacked from the ``Result`` returned by ``check``.
    """

    if not isinstance(tick, int):
        raise GlueError('Tick must be an integer')

    if tick < 0:
        raise GlueError('Tick must be a positive integer')

    logger = logger or Logging.get_logger()

    if timeout is not None:
        end_time = time.time() + timeout

    def _timeout():
        # type: () -> str

        return '{} seconds'.format(int(end_time - time.time())) if timeout is not None else 'infinite'

    logger.debug("waiting for condition '{}', timeout {}, check every {} seconds".format(label, _timeout(),
                                                                                         tick))

    while timeout is None or time.time() < end_time:
        logger.debug("calling callback function")

        check_result = check()

        if check_result.is_ok:
            logger.debug('check passed, assuming success')

            return check_result.unwrap()

        logger.debug("check failed with '{}', assuming failure".format(check_result.value))

        logger.debug('{} left, sleeping for {} seconds'.format(_timeout(), tick))
        time.sleep(tick)

    raise GlueError("Condition '{}' failed to pass within given time".format(label))


def new_xml_element(tag_name, _parent=None, **attrs):
    # type: (str, Optional[Any], **str) -> Any

    """
    Create new XML element.

    :param text tag_name: Name of the element.
    :param element _parent: If set, the newly created element will be appended to this element.
    :param dict attrs: Attributes to set on the newly created element.
    :returns: Newly created XML element.
    """

    element = bs4.BeautifulSoup('', 'xml').new_tag(tag_name)

    for name, value in iteritems(attrs):
        element[name] = value

    if _parent is not None:
        _parent.append(element)

    return element



import string
import types

import pytest
import six

from mock import MagicMock
from hypothesis import example, given, strategies as st

import gluetool
from gluetool import GlueError


@given(message=st.text(string.printable))
def test_message(message):
    with pytest.raises(GlueError) as excinfo:
        raise GlueError(message)

    assert str(excinfo.value) == message


# simulate dummy sys.exc_info values, and add the special "no exception" case,
# like the one reported by sys.exc_info()
@given(caused_by=st.tuples(st.integers(), st.integers(), st.integers()))
@example((None, None, None))
def test_caused_by_explicit(caused_by):
    expected = None if caused_by == (None, None, None) else caused_by

    assert GlueError('dummy message', caused_by=caused_by).caused_by == expected


# without any cause, expect None
def test_caused_by_detect_empty():
    with pytest.raises(GlueError) as excinfo:
        raise GlueError('')

    assert excinfo.value.caused_by is None


# with a cause, find out necessary details
def test_caused_by_detect():
    mock_exc = ValueError('dummy error')

    with pytest.raises(GlueError, match=r'dummy message') as excinfo:
        try:
            raise mock_exc

        except ValueError:
            raise GlueError('dummy message')

    exc = excinfo.value

    assert isinstance(exc.caused_by, tuple)
    assert len(exc.caused_by) == 3

    klass, error, tb = exc.caused_by

    assert klass is mock_exc.__class__
    assert error is mock_exc
    assert isinstance(tb, types.TracebackType)


# sentry_fingerprint works with list(str)
@given(current=st.lists(st.text(string.printable)))
def test_sentry_fingerprint(current):
    assert GlueError('').sentry_fingerprint(current) == current

# sentry_fingerprint argument must be propagated
@given(current=st.lists(st.text(string.printable)), explicit=st.lists(st.text(string.printable)))
def test_sentry_fingerprint(current, explicit):
    if not explicit:
        assert GlueError('', sentry_fingerprint=explicit).sentry_fingerprint(current) is current

    else:
        assert GlueError('', sentry_fingerprint=explicit).sentry_fingerprint(current) == ['GlueError'] + explicit


# sentry_tags works with dict(str, str)
@given(current=st.dictionaries(st.text(string.printable), st.text(string.printable)))
def test_sentry_tags(current):
    assert GlueError('').sentry_tags(current) == current


# sentry_tags argument must be propagated
@given(current=st.dictionaries(st.text(string.printable), st.text(string.printable)),
       explicit=st.dictionaries(st.text(string.printable), st.text(string.printable)))
@example({'foo': 'bar'}, None)
def test_sentry_tags(current, explicit):
    expected = current.copy()
    if explicit:
        expected.update(explicit)

    assert GlueError('', sentry_tags=explicit).sentry_tags(current) == expected


@given(cmd=st.lists(st.text(string.printable)), exit_code=st.integers())
def test_command_error(cmd, exit_code):
    stringified = [six.ensure_str(s) for s in cmd]

    mock_output = MagicMock(exit_code=exit_code)

    exc = gluetool.GlueCommandError(stringified, mock_output)

    assert isinstance(exc, GlueError)
    assert exc.cmd == cmd
    assert exc.output == mock_output
    assert six.ensure_str(six.text_type(exc)) == "Command '{}' failed with exit code {}".format(
        stringified, exit_code
    )


# Simulate module by a simple integer - it's not touched by Failure, therefore it's quite fine.
# exc_info is simulated by tuples of strings - we just want to get them assigned to Failure
# attributes - and to check decisions based on real objects we add few special examples.
@given(module=st.integers(),
       exc_info=st.tuples(st.text(string.printable), st.text(string.printable), st.text(string.printable)))
@example(module=0, exc_info=None)
@example(module=0, exc_info=(None, None, None))
@example(module=0, exc_info=(None, gluetool.SoftGlueError(''), None))
def test_failure(module, exc_info):
    failure = gluetool.Failure(module, exc_info)

    assert failure.module == module
    assert failure.exc_info == exc_info

    assert failure.sentry_event_id is failure.sentry_event_url is None

    if exc_info:
        assert failure.exception == exc_info[1]
        assert failure.soft == isinstance(exc_info[1], gluetool.SoftGlueError)

    else:
        assert failure.exception is None
        assert failure.soft is False



"""
Command-line ``--help`` helpers (muhaha!).

``gluetool`` uses docstrings to generate help for command-line options, modules, shared functions
and other stuff. To generate good looking and useful help texts a bit of work is required.
Add the Sphinx which we use to generate nice documentation of ``gluetool``'s API and structures,
with its directives, and it's even more work to make readable output. Therefore these helpers,
separated in their own file to keep things clean.
"""

import argparse
import ast
import inspect
import os
import textwrap

import docutils.core
import docutils.nodes
import docutils.parsers.rst
import docutils.writers
import sphinx.writers.text
import sphinx.locale
import sphinx.util.nodes

import six
from six import PY2, ensure_str, iteritems

from .color import Colors
from .log import Logging

# Type annotations
# pylint: disable=unused-import, wrong-import-order
from typing import TYPE_CHECKING, cast, Any, Callable, Dict, List, Optional, Tuple, Union  # noqa

if TYPE_CHECKING:
    import gluetool  # noqa
    import gluetool.glue  # noqa


# Initialize Sphinx locale settings
sphinx.locale.init([os.path.split(sphinx.locale.__file__)], None)

# If not told otherwise, the default maximal length of lines is this many columns.
DEFAULT_WIDTH = 120

# Check what's the actual width of our terminal, or use a default if we're not sure.
try:
    WIDTH = int(os.environ['COLUMNS'])

except (KeyError, ValueError):
    WIDTH = DEFAULT_WIDTH

# Crop the maximal width to account for various explicit indents.
CROP_WIDTH = WIDTH - 10

# Tell Sphinx to render text into a slightly narrower space to account for some indenting
sphinx.writers.text.MAXWIDTH = CROP_WIDTH


FUNCTIONS_HELP_TEMPLATE = """
{% for signature, body in FUNCTIONS %}
  {{ signature }}

{{ body }}
{% endfor %}
"""


EVAL_CONTEXT_HELP_TEMPLATE = """
{{ '** Evaluation context **' | style(fg='yellow') }}

{% for name, description in iteritems(CONTEXT) %}
  * {{ name | style(fg='blue') }}

{{ description | indent(4, true) }}
{% endfor %}
"""


# Semantic colorizers
# pylint: disable=invalid-name
def C_FUNCNAME(text):
    # type: (str) -> str

    return Colors.style(text, fg='blue', reset=True)


def C_ARGNAME(text):
    # type: (str) -> str

    return Colors.style(text, fg='blue', reset=True)


def C_LITERAL(text):
    # type: (str) -> str

    return Colors.style(text, fg='cyan', reset=True)


# Our custom TextTranslator which does the same as Sphinx' original but colorizes some of the text bits.
#
# We must save a reference to the original class because we must use it when calling parent's __init__,
# since we cannot use "sphinx.writers.text.TextTranslator" - when we try to call
# sphinx.writers.text.TextTranslator.__init__, it's already set to our custom class => recursion...

# pylint: disable=invalid-name
_original_TextTranslator = sphinx.writers.text.TextTranslator


# pylint: disable=abstract-method
class TextTranslator(sphinx.writers.text.TextTranslator):  # type: ignore  # no type info in TextTranslator
    # literals, ``foo``
    def visit_literal(self, node):
        # type: (Any) -> None

        self.add_text(Colors.style('', fg='cyan', reset=False))

    def depart_literal(self, node):
        # type: (Any) -> None

        self.add_text(Colors.style('', reset=True))

    # "fields" are used to represent (shared) function parameters
    def visit_field_name(self, node):
        # type: (Any) -> None

        _original_TextTranslator.visit_field_name(self, node)

        self.add_text(Colors.style('', fg='blue', reset=False))

    def depart_field_name(self, node):
        # type: (Any) -> None

        self.add_text(Colors.style('', reset=True))

        _original_TextTranslator.depart_field_name(self, node)


sphinx.writers.text.TextTranslator = TextTranslator


# Custom help formatter that let's us control line length
class LineWrapRawTextHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def __init__(self, *args, **kwargs):
        # type: (*Any, **Any) -> None

        if kwargs.get('width', None) is None:
            kwargs['width'] = CROP_WIDTH

        super(LineWrapRawTextHelpFormatter, self).__init__(*args, **kwargs)

    def _split_lines(self, text, width):  # type: ignore  # incompatible with super type because of unicode
        # type: (str, int) -> List[str]

        text = ensure_str(self._whitespace_matcher.sub(' ', ensure_str(text)).strip())

        return textwrap.wrap(text, width)


#
# Code to use Sphinx TextWriter & few our helpers to parse
# our docstring to a plain text.
#

def py_default_role(role, rawtext, text, lineno, inliner, options=None, content=None):
    # type: (Any, str, str, int, Any, Optional[Any], Optional[Any]) -> Tuple[Any, Any]

    # pylint: disable=unused-argument,too-many-arguments
    """
    Default handler we use for ``py:...`` roles, translates text to literal node.
    """

    return [docutils.nodes.literal(rawsource=rawtext, text='{}'.format(text))], []


# register default handler for roles we're interested in
for python_role in ('py:class', 'py:meth', 'py:mod'):
    docutils.parsers.rst.roles.register_canonical_role(python_role, py_default_role)


def doc_role_handler(role, rawtext, text, lineno, inliner, options=None, context=None):
    # type: (Any, str, str, int, Any, Optional[Any], Optional[Any]) -> Tuple[Any, Any]

    # pylint: disable=unused-argument,too-many-arguments
    """
    Format ``:doc:`` roles, used to reference another bits of documentation.
    """

    _, title, target = sphinx.util.nodes.split_explicit_title(text)

    if target and target[0] == '/':
        target = 'docs/source/{}.rst'.format(target[1:])

    return [docutils.nodes.literal(rawsource=text, text='{} (See {})'.format(title, target))], []


docutils.parsers.rst.roles.register_canonical_role('doc', doc_role_handler)


class DummyTextBuilder:
    # pylint: disable=too-few-public-methods,old-style-class,no-init

    """
    Sphinx ``TextWriter`` (and other writers as well) requires an instance of ``Builder``
    class that brings configuration into the rendering process. The original ``TextBuilder``
    requires Sphinx `application` which brings a lot of other dependencies (e.g. source paths
    and similar stuff) which are impractical in our use case ("render short string to plain
    text"). Therefore this dummy class which just provides minimal configuration - ``TextWriter``
    requires nothing else from ``Builder`` instance.

    See ``sphinx/writers/text.py`` for the original implementation.
    """

    class DummyConfig:
        # pylint: disable=too-few-public-methods,old-style-class,no-init

        text_newlines = '\n'
        text_sectionchars = '*=-~"+`'

    config = DummyConfig
    translator_class = None


def rst_to_text(text):
    # type: (str) -> str

    """
    Render given text, written with RST, as plain text.

    :param str text: string to render.
    :rtype: str
    :returns: plain text representation of ``text``.
    """

    return ensure_str(docutils.core.publish_string(text, writer=sphinx.writers.text.TextWriter(DummyTextBuilder)))


def trim_docstring(docstring):
    # type: (str) -> str

    """
    Quoting `PEP 257 <https://www.python.org/dev/peps/pep-0257/#handling-docstring-indentation>`:

    *Docstring processing tools will strip a uniform amount of indentation from
    the second and further lines of the docstring, equal to the minimum indentation
    of all non-blank lines after the first line. Any indentation in the first line
    of the docstring (i.e., up to the first newline) is insignificant and removed.
    Relative indentation of later lines in the docstring is retained. Blank lines
    should be removed from the beginning and end of the docstring.*

    Code bellow follows the quote.

    This method does exactly that, therefore we can keep properly aligned docstrings
    while still use them for reasonably formatted help texts.

    :param str docstring: raw docstring.
    :rtype: str
    :returns: docstring with lines stripped of leading whitespace.
    """

    if not docstring:
        return ''
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = six.MAXSIZE
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < six.MAXSIZE:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return '\n'.join(trimmed)


def docstring_to_help(docstring, width=None, line_prefix='    '):
    # type: (str, Optional[int], str) -> str

    """
    Given docstring, process and render it as a plain text. This conversion function
    is used to generate nice and readable help strings used when printing help on
    command line.

    :param str docstring: raw docstring of Python object (function, method, module, etc.).
    :param int width: Maximal line width allowed.
    :param str line_prefix: prefix each line with this string (e.g. to indent it with few spaces or tabs).
    :returns: formatted docstring.
    """

    width = width or WIDTH

    # Remove leading whitespace
    trimmed = trim_docstring(docstring)

    # we're using Sphinx RST which doesn't look very nice - convert it to plain text then.
    processed = rst_to_text(trimmed)

    # For each line - which is actually a paragraph, given the text comes from RST - wrap it
    # to fit inside given line length (a bit shorter, there's a prefix for each line!).
    wrapped_lines = []  # type: List[str]

    for line in processed.splitlines():
        if line:
            wrapped_lines += textwrap.wrap(
                line,
                width=width - len(line_prefix),
                initial_indent=line_prefix,
                subsequent_indent=line_prefix
            )

        else:
            # yeah, we could just append empty string but line_prefix could be any string, e.g. 'foo: '
            wrapped_lines.append(line_prefix)

    return '\n'.join(wrapped_lines)


def option_help(txt):
    # type: (str) -> str

    """
    Given option help text, format it to be more suitable for command-line help.
    Options can provide a single line of text, or mutiple lines (using triple
    quotes and docstring-like indentation).

    :param str txt: Raw option help text.
    :returns: Formatted option help text.
    """

    # Remove leading whitespace - this won't hurt anyone, and helps docstring-like texts
    trimmed = trim_docstring(txt)

    processed = rst_to_text(trimmed)

    # Merge all lines into a single line
    return ' '.join(processed.splitlines())


def function_help(func, name=None):
    # type: (Callable[..., Any], Optional[str]) -> Tuple[str, str]

    """
    Uses function's signature and docstring to generate a plain text help describing
    the function.

    :param callable func: Function to generate help for.
    :param str name: If not set, ``func.__name__`` is used by default.
    :returns: ``(signature, body)`` pair.
    """

    name = name or func.__name__

    # construct function signature
    # with Python 3 use getfullargspec instead of getargspec
    if PY2:
        # pylint: disable=deprecated-method
        signature = inspect.getargspec(func)
    else:
        signature = inspect.getfullargspec(func)  # pylint: disable=no-member

    no_default = object()

    defaults = []  # type: List[Union[str, object]]

    # arguments that don't have default value are assigned our special value to let us tell the difference
    # between "no default" and "None is the default"
    if signature.defaults is None:
        defaults = [no_default for _ in range(len(signature.args) - 1)]

    else:
        defaults = [
            no_default for _ in range(len(signature.args) - 1 - len(signature.defaults))
        ] + list(signature.defaults)

    args = []
    for arg, default in zip(signature.args[1:], defaults):
        if default is no_default:
            args.append(C_ARGNAME(arg))

        else:
            if isinstance(default, six.string_types):
                default = "'{}'".format(default)

            args.append('{}={}'.format(C_ARGNAME(arg), C_LITERAL(cast(str, default))))

    return (
        # signature
        '{}({})'.format(C_FUNCNAME(name), ', '.join(args)),
        # body
        docstring_to_help(func.__doc__) if func.__doc__ else Colors.style('    No help provided :(', fg='red')
    )


def functions_help(functions):
    # type: (List[Tuple[str, Callable[..., Any]]]) -> str

    """
    Generate help for a set of functions.

    :param list(str, callable) functions: Functions to generate help for, passed as name
        and the corresponding callable pairs.
    :rtype: str
    :returns: Formatted help.
    """

    # pylint: disable=cyclic-import
    from .utils import render_template

    return render_template(
        FUNCTIONS_HELP_TEMPLATE,
        FUNCTIONS=[
            function_help(func, name=name)
            for name, func in functions
        ]
    )


def extract_eval_context_info(source, logger=None):
    # type: (gluetool.glue.Configurable, Optional[gluetool.log.ContextAdapter]) -> Dict[str, str]

    """
    Extract information of evaluation context content from the ``source`` - a module
    or any other object with ``eval_context`` property. The information we're looking
    for is represented by an assignment to special variable, ``__content__``, in the
    body of ``eval_context`` getter. ``__content__`` is expected to be assigned
    a dictionary, listing context variables (keys) and their descriptions (values).

    If it's not possible to find such information, or an exception is raised, the function
    returns an empty dictionary.

    :param gluetool.glue.Configurable source: object to extract information from.
    :rtype: dict(str, str)
    """

    logger = logger or Logging.get_logger()

    logger.debug("extract eval cotext info from '{}'".format(source.name))

    # Cannot do "source.eval_context" because we'd get the value of property, which
    # is usualy a dict. We cannot let it evaluate and return the value, therefore
    # we must get it via its parent class.
    eval_context = source.__class__.eval_context

    # this is not a cyclic import, yet pylint thinks so :/
    # pylint: disable=cyclic-import
    from .glue import Configurable

    # If eval context is the same as Configurable's - which is the original class for all modules
    # and even gluetool.glue.Glue - it means its inherited from Configurable, therefore
    # ``source`` does not have its own, and in that case we should return an empty dictionary,
    # saying "module does not provide any eval context, just the one inherited from the base class",
    # and that's empty anyway.
    if eval_context == Configurable.eval_context:
        logger.debug('eval context matches the original one, ignoring')
        return {}

    try:
        # get source code of the actual getter of the ``eval_context`` property
        getter_source = inspect.getsource(eval_context.fget)  # type: ignore  # the actual property does have `fget`

        # it's indented - trim it like a docstring
        getter_source_trimmed = trim_docstring(getter_source)

        # now, parse getter source, and create its AST
        tree = ast.parse(getter_source_trimmed)

        # find ``__content__ = { ...`` assignment inside the function
        # ``tree`` is the whole module, ``tree.body[0]`` is the function definition
        for node in tree.body[0].body:  # type: ignore  # if `body` is missing, we can handle it
            if not isinstance(node, ast.Assign):
                continue

            target = node.targets[0]

            if not isinstance(target, ast.Name) or target.id != '__content__':
                continue

            if not isinstance(node.value, ast.Dict):
                continue

            assign = node
            break

        else:
            # No "__content__ = {..." found? So be it, return empty info.
            logger.debug('eval context exists but does not describe its content')
            return {}

        # wrap this assignment into a dummy Module node, to create an execution unit with just a single
        # statement (__content__ assignment), so we could slip it to compile/eval.
        dummy_module = ast.Module([assign])

        # compile the module to an executable code
        code = compile(dummy_module, '', 'exec')

        # We prepare our "locals" mapping - when we eval our dummy module, its "__content__ = ..." will be executed
        # within a context of some globals/locals mappings. We give eval our custom locals mapping, which will
        # result in __content__ being created in it - and we can just pick it up from this mapping when eval
        # is done.
        module_locals = {}  # type: Dict[str, Any]

        # this should be reasonably safe, don't raise a warning then...
        # pylint: disable=eval-used
        eval(code, {}, module_locals)

        return {
            name: trim_docstring(description) for name, description in iteritems(module_locals['__content__'])
        }

    # pylint: disable=broad-except
    except Exception as exc:
        logger.warning("Cannot read eval context info from '{}': {}".format(source.name, exc))

        return {}


def eval_context_help(source):
    # type: (gluetool.glue.Configurable) -> str

    """
    Generate and format help for an evaluation context of a module. Looks for context content,
    and gives it a nice header, suitable for command-line help, applying formatting along the way.

    :param gluetool.glue.Configurable source: object whose eval context help we should format.
    :returns: Formatted help.
    """

    # pylint: disable=cyclic-import
    from .utils import render_template

    context_info = extract_eval_context_info(source)

    if not context_info:
        return ''

    context_content = {
        name: docstring_to_help(description, line_prefix='') for name, description in iteritems(context_info)
    }

    return render_template(EVAL_CONTEXT_HELP_TEMPLATE, CONTEXT=context_content)



# pylint: disable=blacklisted-name

import logging
import pytest

import gluetool
import pkg_resources

from mock import MagicMock


class DummyModule(gluetool.Module):
    name = 'dummy-module'


class DummyModuleWithAliases(gluetool.Module):
    name = ('dummy-module', 'dummy-module-alias')


class DummyModuleWithoutName(gluetool.Module):
    pass


@pytest.fixture(name='glue')
def fixture_glue():
    return gluetool.Glue()


def _assert_module_registered(log, glue, registry, klass, name, group):
    assert name in registry
    assert registry[name].klass is klass
    assert registry[name].group == group

    assert log.match(message="registering module '{}' from dummy-filepath:{}".format(name, klass.__name__))


def test_register_module(log, glue):
    registry = {}

    glue._register_module(registry, 'dummy-group', DummyModule, 'dummy-filepath')

    assert len(registry) == 1

    _assert_module_registered(log, glue, registry, DummyModule, 'dummy-module', 'dummy-group')


def test_register_module_aliases(log, glue):
    registry = {}

    glue._register_module(registry, 'dummy-group', DummyModuleWithAliases, 'dummy-filepath')

    assert len(registry) == 2

    _assert_module_registered(log, glue, registry, DummyModuleWithAliases, 'dummy-module', 'dummy-group')
    _assert_module_registered(log, glue, registry, DummyModuleWithAliases, 'dummy-module-alias', 'dummy-group')


def test_register_module_no_names(log, glue):
    registry = {}

    with pytest.raises(gluetool.GlueError, match=r'No name specified by module class dummy-filepath:DummyModule'):
        glue._register_module(registry, 'dummy-group', DummyModuleWithoutName, 'dummy-filepath')


def test_register_module_name_conflict(log, glue):
    registry = {
        'dummy-module': None
    }

    with pytest.raises(gluetool.GlueError, match=r"Name 'dummy-module' of class dummy-filepath:DummyModule is a duplicate module name"):
        glue._register_module(registry, 'dummy-group', DummyModule, 'dummy-filepath')


def test_discover_gm_in_entry(log, monkeypatch, glue):
    registry = {}

    mock_ep = MagicMock(load=MagicMock(return_value=DummyModule), dist=MagicMock(location='dummy-filepath'))
    mock_iter_entry_points = MagicMock(return_value=[mock_ep])

    monkeypatch.setattr(pkg_resources, 'iter_entry_points', mock_iter_entry_points)

    glue._discover_gm_in_entry_point('dummy-entry-point', registry)

    mock_ep.load.assert_called_once()

    assert len(registry) == 1
    _assert_module_registered(log, glue, registry, DummyModule, 'dummy-module', '')


@pytest.mark.parametrize(
    ('expected', 'error_message', 'content'),
    (
        # Empty Python file
        (False, "  no 'import gluetool' found", 'pass'),

        # Check file that imports gluetool but does not have module class
        (False, '  no child of gluetool.Module found', 'import gluetool'),
        (False, '  no child of gluetool.Module found', 'from gluetool import Glue'),

        # Check we ignore module classes with wrong base class
        (False, '  no child of gluetool.Module found', """
import gluetool

class DummyModule(object):
    pass
"""),

        # Check file that does have module class, but that does not import gluetool
        (False, "  no 'import gluetool' found", """
class DummyModule(gluetool.Module):
    pass
"""),

        # Check file that both imports gluetool, and has module class
        (True, '', """
import gluetool

class DummyModule(gluetool.Module):
    pass
"""),

        (True, '', """
from gluetool import Module

class DummyModule(Module):
    pass
"""),

        # Check if module can have multiple names
        (True, '', """
from gluetool import Module

class DummyModule(Module):
    name = ['dummy', 'dummy-alias']
""")
    )
)
def test_check_pm_file(log, tmpdir, glue, expected, error_message, content):
    # pylint: disable=protected-access

    pm_file = tmpdir.join('dummy.py')
    pm_file.write(content)

    log.clear()

    assert glue._check_pm_file(str(pm_file)) is expected

    assert log.match(levelno=logging.DEBUG, message="check possible module file '{}'".format(pm_file))

    if expected is False:
        assert log.match(levelno=logging.DEBUG, message=error_message)


def test_check_pm_file_missing(log, tmpdir, glue):
    with pytest.raises(gluetool.GlueError, match=r"Unable to check check module file 'foo\.txt': \[Errno 2\] No such file or directory: 'foo\.txt'"):
        glue._check_pm_file('foo.txt')



import pytest

import gluetool
from gluetool.result import Result, Ok, Error


@pytest.mark.parametrize('result, value, is_ok', [
    (Ok(1), 1, True),
    (Result.Ok(1), 1, True),
    (Ok(True), True, True),
    (Ok(False), False, True),
    (Error(1), 1, False),
    (Result.Error(1), 1, False),
    (Error(True), True, False)
])
def test_factory(result, value, is_ok):
    assert result._value == value
    assert result.is_ok is is_ok
    assert result.is_error is not is_ok


def test_eq():
    assert Ok(1) == Ok(1)
    assert Error(1) == Error(1)
    assert Ok(1) != Error(1)
    assert Ok(1) != Ok(2)
    assert not (Ok(1) != Ok(1))
    assert Ok(1) != 'foo'
    assert Ok('0') != Ok(0)


def test_hash():
    assert len({Ok(1), Error('2'), Ok(1), Error('2')}) == 2
    assert len({Ok(1), Ok(2)}) == 2
    assert len({Ok('a'), Error('a')}) == 2


def test_ok_value():
    o = Ok('foo')
    n = Error('foo')
    assert o.ok == 'foo'
    assert n.ok is None


def test_err_value():
    o = Ok('foo')
    n = Error('foo')
    assert o.error is None
    assert n.error == 'foo'


def test_no_constructor():
    with pytest.raises(RuntimeError):
        Result(_is_ok=True, _value='foo')


def test_unwrap():
    o = Ok('foo')
    n = Error('foo')

    assert o.unwrap() == 'foo'

    with pytest.raises(gluetool.GlueError):
        n.unwrap()


def test_expect():
    o = Ok('foo')
    n = Error('foo')

    assert o.expect('failure') == 'foo'

    with pytest.raises(gluetool.GlueError):
        n.expect('failure')


def test_unwrap_or():
    o = Ok('foo')
    n = Error('foo')

    assert o.unwrap_or('default value') == 'foo'
    assert n.unwrap_or('default value') == 'default value'




"""
Heart of the "gluetool" script. Referred to by setuptools' entry point.
"""

from __future__ import print_function

import functools
import logging
import os
import re
import signal
import sys
import traceback

from six import ensure_str, iteritems, iterkeys

import tabulate

import gluetool
import gluetool.action
import gluetool.sentry

from .glue import GlueError, GlueRetryError, Failure, PipelineStepModule
from .help import extract_eval_context_info, docstring_to_help
from .log import log_dict
from .utils import format_command_line, cached_property, normalize_path, render_template, normalize_multistring_option

# Type annotations
# pylint: disable=unused-import,wrong-import-order
from typing import cast, Any, Callable, List, Optional, NoReturn, Union  # noqa
from types import FrameType  # noqa
from gluetool.glue import PipelineReturnType  # noqa


# Order is important, the later one overrides values from the former
DEFAULT_GLUETOOL_CONFIG_PATHS = [
    '/etc/gluetool.d/gluetool',
    normalize_path('~/.gluetool.d/gluetool'),
    normalize_path('./.gluetool.d/gluetool')
]

DEFAULT_HANDLED_SIGNALS = (signal.SIGUSR2,)


def handle_exc(func):
    # type: (Callable[..., Any]) -> Callable[..., Any]

    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        # type: (Gluetool, *Any, **Any) -> Any

        # pylint: disable=broad-except, protected-access

        try:
            return func(self, *args, **kwargs)

        except (SystemExit, KeyboardInterrupt, Exception):
            self._handle_failure(Failure(self.Glue.current_module if self.Glue else None, sys.exc_info()))

    return wrapped


class Gluetool(object):
    def __init__(self):
        # type: () -> None

        self.gluetool_config_paths = DEFAULT_GLUETOOL_CONFIG_PATHS

        self.sentry = None  # type: Optional[gluetool.sentry.Sentry]
        self.tracer = None  # type: Optional[gluetool.action.Tracer]

        # pylint: disable=invalid-name
        self.Glue = None  # type: Optional[gluetool.glue.Glue]

        self.argv = None  # type: Optional[List[str]]
        self.pipeline_desc = None  # type: Optional[List[gluetool.glue.PipelineStepModule]]

    @cached_property
    def _version(self):
        # type: () -> str

        # pylint: disable=no-self-use
        from .version import __version__

        return ensure_str(__version__.strip())

    @cached_property
    def _command_name(self):
        # type: () -> str

        # pylint: disable=no-self-use
        return 'gluetool'

    def _deduce_pipeline_desc(self, argv, modules):
        # type: (List[Any], List[str]) -> List[gluetool.glue.PipelineStepModule]
        # pylint: disable=no-self-use

        """
        Split command-line arguments, left by ``gluetool``, into a pipeline description, splitting them
        by modules and their options.

        :param list argv: Remainder of :py:data:`sys.argv` after removing ``gluetool``'s own options.
        :param list(text) modules: List of known module names.
        :returns: Pipeline description in a form of a list of :py:class:`gluetool.glue.PipelineStepModule` instances.
        """

        alias_pattern = re.compile(r'^([a-z\-]*):([a-z\-]*)$', re.I)

        pipeline_desc = []
        step = None

        while argv:
            arg = argv.pop(0)

            # is the "arg" a module name? If so, add new step to the pipeline
            if arg in modules:
                step = PipelineStepModule(arg)
                pipeline_desc.append(step)
                continue

            # is the "arg" a module with an alias? If so, add a new step to the pipeline, and note the alias
            match = alias_pattern.match(arg)
            if match is not None:
                module, actual_module = match.groups()

                step = PipelineStepModule(module, actual_module=actual_module)
                pipeline_desc.append(step)
                continue

            if step is None:
                raise GlueError("Cannot parse module argument: '{}'".format(arg))

            step.argv.append(arg)

        return pipeline_desc

    def log_cmdline(self, argv, pipeline_desc):
        # type: (List[Any], List[gluetool.glue.PipelineStepModule]) -> None

        cmdline = [
            [ensure_str(sys.argv[0])] + argv
        ]

        for step in pipeline_desc:
            cmdline.append([step.module_designation] + step.argv)

        assert self.Glue is not None
        self.Glue.info('command-line:\n{}'.format(format_command_line(cmdline)))

    @cached_property
    def _exit_logger(self):
        # type: () -> Union[logging.Logger, gluetool.log.ContextAdapter]

        # pylint: disable=no-self-use
        """
        Return logger for use when finishing the ``gluetool`` pipeline.
        """

        # We want to use the current logger, if there's any set up.
        logger = gluetool.log.Logging.get_logger()  # type: Union[logging.Logger, gluetool.log.ContextAdapter]

        if logger:
            return logger

        # This may happen only when something went wrong during logger initialization
        # when Glue instance was created. Falling back to a very basic Logger seems
        # to be the best option here.

        logging.basicConfig(level=logging.DEBUG)
        logger = logging.getLogger()

        logger.warning('Cannot use custom logger, falling back to a default one')

        return logger

    def _quit(self, exit_status):
        # type: (int) -> NoReturn

        """
        Log exit status and quit.
        """

        logger = self._exit_logger

        if self.tracer:
            self.tracer.close(logger=logger)

        (logger.debug if exit_status == 0 else logger.error)('Exiting with status {}'.format(exit_status))

        sys.exit(exit_status)

    # pylint: disable=invalid-name
    def _handle_failure_core(self, failure):
        # type: (gluetool.glue.Failure) -> NoReturn

        logger = self._exit_logger

        assert failure.exc_info is not None
        assert failure.exc_info[1] is not None

        # Handle simple 'sys.exit(0)' - no exception happened
        if failure.exc_info[0] == SystemExit:
            assert isinstance(failure.exc_info[1], SystemExit)  # collapse type to SystemExit to make mypy happy

            if failure.exc_info[1].code == 0:
                self._quit(0)

        # soft errors are up to users to fix, no reason to kill pipeline
        exit_status = 0 if failure.soft is True else -1

        if failure.module:
            msg = "Pipeline reported an exception in module '{}': {}".format(
                failure.module.unique_name,
                failure.exc_info[1]
            )

        else:
            msg = "Pipeline reported an exception: {}".format(failure.exc_info[1])

        logger.error(msg, exc_info=failure.exc_info)

        # Submit what hasn't been submitted yet...
        if self.sentry and not failure.sentry_event_id:
            self.sentry.submit_exception(failure, logger=logger)

        self._quit(exit_status)

    # pylint: disable=invalid-name
    def _handle_failure(self, failure):
        # type: (gluetool.glue.Failure) -> NoReturn

        try:
            self._handle_failure_core(failure)

        # pylint: disable=broad-except
        except Exception:
            exc_info = sys.exc_info()

            # Don't trust anyone, the exception might have occured inside logging code, therefore
            # resorting to plain print.

            err_print = functools.partial(print, file=sys.stderr)

            err_print("""
!!! While handling an exception, another one appeared !!!

Will try to submit it to Sentry but giving up on everything else.
""")

            try:
                # pylint: disable=protected-access
                err_print(gluetool.log.LoggingFormatter._format_exception_chain(sys.exc_info()))

                # Anyway, try to submit this exception to Sentry, but be prepared for failure in case the original
                # exception was raised right in Sentry-related code.
                if self.sentry is not None:
                    self.sentry.submit_exception(Failure(None, exc_info))

            # pylint: disable=broad-except
            except Exception:
                # tripple error \o/

                err_print("""
!!! While submitting an exception to the Sentry, yet another exception appeared !!!
    Giving up on everything...
""")

                traceback.print_exc()

            # Don't use _quit() here - it might try to use complicated logger, and we don't trust
            # anythign at this point. Just die already.
            sys.exit(-1)

    @handle_exc
    def setup(self):
        # type: () -> None

        self.sentry = gluetool.sentry.Sentry()
        self.tracer = gluetool.action.Tracer()

        # Python installs SIGINT handler that translates signal to
        # a KeyboardInterrupt exception. It's so good we want to use
        # it for SIGTERM as well, just wrap the handler with some logging.
        orig_sigint_handler = signal.getsignal(signal.SIGINT)
        sigmap = {getattr(signal, name): name for name in [name for name in dir(signal) if name.startswith('SIG')]}

        def _signal_handler(signum, frame, handler=None, msg=None):
            # type: (int, FrameType, Optional[Callable[[int, FrameType], None]], Optional[str]) -> Any

            msg = msg or 'Signal {} received'.format(sigmap[signum])

            Glue.warn(msg)

            if handler is not None:
                return handler(signum, frame)

        def _sigusr1_handler(signum, frame):
            # type: (int, FrameType) -> None

            # pylint: disable=unused-argument

            raise GlueError('Pipeline timeout expired.')

        sigint_handler = functools.partial(_signal_handler,
                                           handler=orig_sigint_handler, msg='Interrupted by SIGINT (Ctrl+C?)')
        sigterm_handler = functools.partial(_signal_handler,
                                            handler=orig_sigint_handler, msg='Interrupted by SIGTERM')
        sigusr1_handler = functools.partial(_signal_handler, handler=_sigusr1_handler)

        # pylint: disable=invalid-name
        Glue = self.Glue = gluetool.glue.Glue(tool=self, sentry=self.sentry)

        # Glue is initialized, we can install our logging handlers
        signal.signal(signal.SIGINT, sigint_handler)
        signal.signal(signal.SIGTERM, sigterm_handler)
        signal.signal(signal.SIGUSR1, sigusr1_handler)

        for signum in DEFAULT_HANDLED_SIGNALS:
            signal.signal(signum, _signal_handler)

        # process configuration
        Glue.parse_config(self.gluetool_config_paths)
        Glue.parse_args(sys.argv[1:])

        # store tool's configuration - everything till the start of "pipeline" (the first module)
        self.argv = [
            ensure_str(arg) for arg in sys.argv[1:len(sys.argv) - len(Glue.option('pipeline'))]
        ]

        if Glue.option('pid'):
            Glue.info('PID: {} PGID: {}'.format(os.getpid(), os.getpgrp()))

        # version
        if Glue.option('version'):
            Glue.info('{} {}'.format(self._command_name, self._version))
            sys.exit(0)

        GlueError.no_sentry_exceptions = normalize_multistring_option(Glue.option('no-sentry-exceptions'))

        Glue.modules = Glue.discover_modules()

    @handle_exc
    def check_options(self):
        # type: () -> None

        Glue = self.Glue
        assert Glue is not None

        self.pipeline_desc = self._deduce_pipeline_desc(Glue.option('pipeline'), list(iterkeys(Glue.modules)))
        log_dict(Glue.debug, 'pipeline description', self.pipeline_desc)

        # list modules
        groups = Glue.option('list-modules')
        if groups == [True]:
            sys.stdout.write('%s\n' % Glue.modules_descriptions())
            sys.exit(0)

        elif groups:
            sys.stdout.write('%s\n' % Glue.modules_descriptions(groups=groups))
            sys.exit(0)

        if Glue.option('list-shared'):
            functions = []  # type: List[List[str]]

            for mod_name in sorted(iterkeys(Glue.modules)):
                functions += [
                    [func_name, mod_name] for func_name in Glue.modules[mod_name].klass.shared_functions
                ]

            if functions:
                functions = sorted(functions, key=lambda row: row[0])
            else:
                functions = [['-- no shared functions available --', '']]

            sys.stdout.write("""Available shared functions

{}
            """.format(tabulate.tabulate(functions, ['Shared function', 'Module name'], tablefmt='simple')))

            sys.exit(0)

        if Glue.option('list-eval-context'):
            variables = []

            def _add_variables(source):
                # type: (gluetool.glue.Configurable) -> None

                info = extract_eval_context_info(source)

                for name, description in iteritems(info):
                    variables.append([
                        name, source.name, docstring_to_help(description, line_prefix='')
                    ])

            for mod_name in sorted(iterkeys(Glue.modules)):
                _add_variables(Glue.init_module(mod_name))

            _add_variables(Glue)

            if variables:
                variables = sorted(variables, key=lambda row: row[0])

            else:
                variables = [['-- no variables available --', '', '']]

            table = tabulate.tabulate(variables, ['Variable', 'Module name', 'Description'], tablefmt='simple')

            print(render_template("""
{{ '** Variables available in eval context **' | style(fg='yellow') }}

{{ TABLE }}
            """, TABLE=table))

            sys.exit(0)

    @handle_exc
    def run_pipeline(self):
        # type: () -> PipelineReturnType

        Glue = self.Glue
        assert Glue is not None

        # no modules
        if not self.pipeline_desc:
            raise GlueError('No module specified, use -l to list available')

        # command-line info
        if Glue.option('info'):
            assert self.argv is not None

            self.log_cmdline(self.argv, self.pipeline_desc)

        # actually the execution loop is retries+1
        # there is always one execution
        retries = Glue.option('retries')

        for loop_number in range(retries + 1):
            # Print retry info
            if loop_number:
                Glue.warning('retrying execution (attempt #{} out of {})'.format(loop_number, retries))

            # Run the pipeline
            failure, destroy_failure = Glue.run_modules(self.pipeline_desc)

            if destroy_failure:
                return failure, destroy_failure

            if failure and isinstance(failure.exception, GlueRetryError):
                Glue.error(str(failure.exception))
                continue

            return failure, destroy_failure

        return None, None

    def main(self):
        # type: () -> None

        self.setup()
        self.check_options()

        failure, destroy_failure = self.run_pipeline()

        if destroy_failure:
            self._exit_logger.warning('Exception raised when destroying modules, overriding exit status')

            self._handle_failure(destroy_failure)

        if failure:
            self._handle_failure(failure)

        self._quit(0)


def main():
    # type: () -> None

    app = Gluetool()
    app.main()



import os
import string

import pytest
from hypothesis import assume, given, strategies as st

from gluetool.utils import normalize_bool_option, normalize_multistring_option, normalize_path, normalize_path_option, normalize_shell_option


def all_casings(input_string):
    if not input_string:
        yield u''

    else:
        first = input_string[:1]
        for sub_casing in all_casings(input_string[1:]):
            yield first.lower() + sub_casing
            yield first.upper() + sub_casing


# all of these are expected to evaluate to ``True``
POSITIVE_BOOLEAM_OPTIONS = [
    True] \
    + list(all_casings(u'yes')) \
    + list(all_casings(u'true')) \
    + [u'1'] \
    + [u'y', u'Y'] \
    + list(all_casings(u'on'))


# test not just values, but also check it works when they are wrapped by some whitespace
@pytest.mark.parametrize('value',
                         POSITIVE_BOOLEAM_OPTIONS + [u'\t {}\t '.format(value) for value in POSITIVE_BOOLEAM_OPTIONS])
def test_normalize_bool_positive(value):
    assert normalize_bool_option(value) is True


# Anything not listed in POSITIVE_BOOLEAM_OPTIONS should be evaluated to False, including False :)
# Ask hypothesis to run the test with either False or huge pile of weird strings, just make sure
# their stripped version is not in POSITIVE_BOOLEAM_OPTIONS (there's a lot of whitespace in
# string.printable).
@given(value=st.one_of(st.just(False), st.text(alphabet=string.printable)))
def test_normalize_bool_negative(value):
    if not isinstance(value, bool):
        assume(value.strip() not in POSITIVE_BOOLEAM_OPTIONS)

    assert normalize_bool_option(value) is False


def test_normalize_multistring_none():
    assert normalize_multistring_option(None) == []


def test_normalize_multistring_string():
    assert normalize_multistring_option(u'foo') == [u'foo']


# For sake of simplicity, use ASCII lowercase, add comma to get "foo,bar" for free, letting
# us check whether splitting of multiple values in a single item works as well.
@given(value=st.lists(st.text(alphabet=string.ascii_lowercase + ',')))
def test_normalize_multistring(value):
    expected = sum([s.strip().split(',') for s in value], [])

    assert normalize_multistring_option(value) == expected


# Generate path from simple alphabet, throw in slash and tilda to make it look
# like a path.
@given(value=st.text(alphabet=string.ascii_lowercase + '/~'))
def test_normalize_path(value):
    expected = os.path.abspath(os.path.expanduser(value))

    assert normalize_path(value) == expected


# Generate paths, and comma (like test_normalize_multistring) to get multiple paths in a single
# string for free.
@given(value=st.lists(st.text(alphabet=string.ascii_lowercase + '/~,')))
def test_normalize_path_option(value):
    expected = [normalize_path(path) for paths in value for path in paths.split(',')]

    assert normalize_path_option(value) == expected


@pytest.mark.parametrize('option_value, expected', (
    (
        u'--foo --bar',
        [u'--foo', u'--bar']
    ),
    (
        [u'--foo --bar', u'--baz="abc def"', u'--foo abc\\ def'],
        [u'--foo', u'--bar', u'--baz=abc def', u'--foo', u'abc def']
    )
))
def test_normalize_shell_option(option_value, expected):
    assert normalize_shell_option(option_value) == expected



import re
import sys

import jinja2
import pytest

import gluetool
import gluetool.log


EXPECTED = jinja2.Template(r"""---v---v---v---v---v--- Exception: ---v---v---v---v---v---

At {{ FILE }}:81, in foo:

gluetool.glue.GlueError: Foo failed


  File "{{ FILE }}", line 87, in test_sanity
    foo\(\)

    Local variables:
        exc = Foo failed
        excinfo = \(<class 'gluetool\.glue\.GlueError'>, GlueError\('Foo failed',\), <traceback object at 0x[0-9a-f]+>\)

  File "{{ FILE}}", line 81, in foo
    raise gluetool.GlueError\('Foo failed'\)

    Local variables:

---\^---\^---\^---\^---\^---\^----------\^---\^---\^---\^---\^---\^---

---v---v---v---v---v--- Caused by: ---v---v---v---v---v---

At {{ FILE }}:69, in baz:

{{ VALUE_ERROR_MODULE }}.ValueError: Z is really lame value


  File "{{ FILE }}", line 78, in foo
    return bar\(17\)

    Local variables:

  File "{{ FILE }}", line 73, in bar
    return baz\(13, p\)

    Local variables:
        p = 17

  File "{{ FILE }}", line 69, in baz
    raise ValueError\('Z is really lame value'\)

    Local variables:
        x = 13
        y = 17
        z = 13

---\^---\^---\^---\^---\^---\^----------\^---\^---\^---\^---\^---\^---
""").render(FILE=__file__, VALUE_ERROR_MODULE=ValueError.__module__)


# Raise exception from a frame deeper in the stack, catch it and raise another
# to establish exception chain. Spice the functions with local variables, to give
# us something to check in the output.

def baz(x, y):
    z = x

    raise ValueError('Z is really lame value')


def bar(p):
    return baz(13, p)


def foo():
    try:
        return bar(17)

    except ValueError:
        raise gluetool.GlueError('Foo failed')


def test_sanity():
    # not using pytest.raises, don't want pytest to spoil our exception with its smart stuff
    try:
        foo()

    except gluetool.GlueError as exc:
        excinfo = sys.exc_info()

        # not assigning output of formatter to a local variable, we don't want it to appear in the top-level
        # frame's list of locals, leading to a self-reference in expected output.

        # match lines one by one, using expected as a regex pattern
        for l1, l2 in zip(EXPECTED.split('\n'), gluetool.log.LoggingFormatter._format_exception_chain(excinfo).split('\n')):
            assert re.match('^' + l1 + '$', l2)
